// Generated by https://github.com/juliettef/IconFontCppHeaders script GenerateIconFontCppHeaders.py for language C89
// from https://raw.githubusercontent.com/nicodinh/kenney-icon-font/master/css/kenney-icons.css
// for use with https://github.com/nicodinh/kenney-icon-font/blob/master/fonts/kenney-icon-font.ttf
#pragma once

#define FONT_ICON_FILE_NAME_KI "kenney-icon-font.ttf"

#define ICON_MIN_KI 0xe900
#define ICON_MAX_KI 0xe9e3
#define ICON_KI_HOME "\xee\xa4\x80"
#define ICON_KI_ADJUST "\xee\xa4\x81"
#define ICON_KI_WRENCH "\xee\xa4\x82"
#define ICON_KI_COG "\xee\xa4\x83"
#define ICON_KI_OFF "\xee\xa4\x84"
#define ICON_KI_EXPAND "\xee\xa4\x85"
#define ICON_KI_REDUCE "\xee\xa4\x86"
#define ICON_KI_MOVIE "\xee\xa4\x87"
#define ICON_KI_FLAP "\xee\xa4\x88"
#define ICON_KI_SHOPPING_CART "\xee\xa4\x89"
#define ICON_KI_SHOPPING_CASE "\xee\xa4\x8a"
#define ICON_KI_EXTERNAL "\xee\xa4\x8b"
#define ICON_KI_NETWORK "\xee\xa4\x8c"
#define ICON_KI_CHECK "\xee\xa4\x8d"
#define ICON_KI_TIMES "\xee\xa4\x8e"
#define ICON_KI_TIMES_CIRCLE "\xee\xa4\x8f"
#define ICON_KI_PLUS "\xee\xa4\x90"
#define ICON_KI_PLUS_CIRCLE "\xee\xa4\x91"
#define ICON_KI_MINUS "\xee\xa4\x92"
#define ICON_KI_MINUS_CIRCLE "\xee\xa4\x93"
#define ICON_KI_INFO "\xee\xa4\x94"
#define ICON_KI_INFO_CIRCLE "\xee\xa4\x95"
#define ICON_KI_QUESTION "\xee\xa4\x96"
#define ICON_KI_QUESTION_CIRCLE "\xee\xa4\x97"
#define ICON_KI_EXLAMATION "\xee\xa4\x98"
#define ICON_KI_EXCLAMATION_CIRCLE "\xee\xa4\x99"
#define ICON_KI_EXCLAMATION_TRIANGLE "\xee\xa4\x9a"
#define ICON_KI_PAINT_BRUSH "\xee\xa4\x9b"
#define ICON_KI_PENCIL "\xee\xa4\x9c"
#define ICON_KI_CHECKBOX "\xee\xa4\x9d"
#define ICON_KI_CHECKBOX_CHECKED "\xee\xa4\x9e"
#define ICON_KI_RADIO "\xee\xa4\x9f"
#define ICON_KI_RADIO_CHECKED "\xee\xa4\xa0"
#define ICON_KI_SORT_VERTICAL "\xee\xa4\xa1"
#define ICON_KI_SORT_HORIZONTAL "\xee\xa4\xa2"
#define ICON_KI_GRID "\xee\xa4\xa3"
#define ICON_KI_LIST "\xee\xa4\xa4"
#define ICON_KI_ROWS "\xee\xa4\xa5"
#define ICON_KI_CELLS "\xee\xa4\xa6"
#define ICON_KI_SIGNAL_LOW "\xee\xa4\xa7"
#define ICON_KI_SIGNAL_MEDIUM "\xee\xa4\xa8"
#define ICON_KI_SIGNAL_HIGH "\xee\xa4\xa9"
#define ICON_KI_TRASH "\xee\xa4\xaa"
#define ICON_KI_TRASH_ALT "\xee\xa4\xab"
#define ICON_KI_RELOAD_INVERSE "\xee\xa4\xac"
#define ICON_KI_RELOAD "\xee\xa4\xad"
#define ICON_KI_TOP "\xee\xa4\xae"
#define ICON_KI_BOTTOM "\xee\xa4\xaf"
#define ICON_KI_UPLOAD "\xee\xa4\xb0"
#define ICON_KI_DOWNLOAD "\xee\xa4\xb1"
#define ICON_KI_CLOUD "\xee\xa4\xb2"
#define ICON_KI_CLOUD_UPLOAD "\xee\xa4\xb3"
#define ICON_KI_CLOUD_DOWNLOAD "\xee\xa4\xb4"
#define ICON_KI_SEARCH "\xee\xa4\xb5"
#define ICON_KI_SEARCH_PLUS "\xee\xa4\xb6"
#define ICON_KI_SEARCH_MINUS "\xee\xa4\xb7"
#define ICON_KI_SEARCH_EQUAL "\xee\xa4\xb8"
#define ICON_KI_LOCK "\xee\xa4\xb9"
#define ICON_KI_UNLOCK "\xee\xa4\xba"
#define ICON_KI_USER "\xee\xa4\xbb"
#define ICON_KI_USERS "\xee\xa4\xbc"
#define ICON_KI_USERS_ALT "\xee\xa4\xbd"
#define ICON_KI_SIGN_IN "\xee\xa4\xbe"
#define ICON_KI_SIGN_IN_INVERSE "\xee\xa4\xbf"
#define ICON_KI_SIGN_OUT "\xee\xa5\x80"
#define ICON_KI_SIGN_OUT_INVERSE "\xee\xa5\x81"
#define ICON_KI_ARROW_TOP "\xee\xa5\x82"
#define ICON_KI_ARROW_RIGHT "\xee\xa5\x83"
#define ICON_KI_ARROW_BOTTOM "\xee\xa5\x84"
#define ICON_KI_ARROW_LEFT "\xee\xa5\x85"
#define ICON_KI_ARROW_TOP_LEFT "\xee\xa5\x86"
#define ICON_KI_ARROW_TOP_RIGHT "\xee\xa5\x87"
#define ICON_KI_ARROW_BOTTOM_RIGHT "\xee\xa5\x88"
#define ICON_KI_ARROW_BOTTOM_LEFT "\xee\xa5\x89"
#define ICON_KI_CARET_TOP "\xee\xa5\x8a"
#define ICON_KI_CARET_RIGHT "\xee\xa5\x8b"
#define ICON_KI_CARET_BOTTOM "\xee\xa5\x8c"
#define ICON_KI_CARET_LEFT "\xee\xa5\x8d"
#define ICON_KI_NEXT_ALT "\xee\xa5\x8e"
#define ICON_KI_NEXT "\xee\xa5\x8f"
#define ICON_KI_PREVIOUS "\xee\xa5\x90"
#define ICON_KI_PREVIOUS_ALT "\xee\xa5\x91"
#define ICON_KI_FILL "\xee\xa5\x92"
#define ICON_KI_ERASER "\xee\xa5\x93"
#define ICON_KI_SAVE "\xee\xa5\x94"
#define ICON_KI_STEP_BACKWARD "\xee\xa5\x95"
#define ICON_KI_BACKWARD "\xee\xa5\x96"
#define ICON_KI_PAUSE "\xee\xa5\x97"
#define ICON_KI_FORWARD "\xee\xa5\x98"
#define ICON_KI_STEP_FORWARD "\xee\xa5\x99"
#define ICON_KI_STOP "\xee\xa5\x9a"
#define ICON_KI_REC "\xee\xa5\x9b"
#define ICON_KI_CURSOR "\xee\xa5\x9c"
#define ICON_KI_POINTER "\xee\xa5\x9d"
#define ICON_KI_EXIT "\xee\xa5\x9e"
#define ICON_KI_FIGURE "\xee\xa5\x9f"
#define ICON_KI_CAR "\xee\xa5\xa0"
#define ICON_KI_COIN "\xee\xa5\xa1"
#define ICON_KI_KEY "\xee\xa5\xa2"
#define ICON_KI_CUB "\xee\xa5\xa3"
#define ICON_KI_DIAMOND "\xee\xa5\xa4"
#define ICON_KI_BADGE "\xee\xa5\xa5"
#define ICON_KI_BADGE_ALT "\xee\xa5\xa6"
#define ICON_KI_PODIUM "\xee\xa5\xa7"
#define ICON_KI_PODIUM_ALT "\xee\xa5\xa8"
#define ICON_KI_FLAG "\xee\xa5\xa9"
#define ICON_KI_FIST "\xee\xa5\xaa"
#define ICON_KI_FIST_CIRCLE "\xee\xa5\xab"
#define ICON_KI_HEART "\xee\xa5\xac"
#define ICON_KI_HEART_HALF "\xee\xa5\xad"
#define ICON_KI_HEART_HALF_O "\xee\xa5\xae"
#define ICON_KI_HEART_O "\xee\xa5\xaf"
#define ICON_KI_STAR "\xee\xa5\xb0"
#define ICON_KI_STAR_HALF "\xee\xa5\xb1"
#define ICON_KI_STAR_HALF_O "\xee\xa5\xb2"
#define ICON_KI_STAR_O "\xee\xa5\xb3"
#define ICON_KI_BUTTON_B "\xee\xa5\xb4"
#define ICON_KI_MUSIC_ON "\xee\xa5\xb5"
#define ICON_KI_MUSIC_OFF "\xee\xa5\xb6"
#define ICON_KI_SOUND_ON "\xee\xa5\xb7"
#define ICON_KI_SOUND_OFF "\xee\xa5\xb8"
#define ICON_KI_SOUND_OFF_ALT "\xee\xa5\xb9"
#define ICON_KI_ROBOT "\xee\xa5\xba"
#define ICON_KI_COMPUTER "\xee\xa5\xbb"
#define ICON_KI_TABLET "\xee\xa5\xbc"
#define ICON_KI_SMARTPHONE "\xee\xa5\xbd"
#define ICON_KI_DEVICE "\xee\xa5\xbe"
#define ICON_KI_DEVICE_TILT_LEFT "\xee\xa5\xbf"
#define ICON_KI_DEVICE_TILT_RIGHT "\xee\xa6\x80"
#define ICON_KI_GAMEPAD "\xee\xa6\x81"
#define ICON_KI_GAMEPAD_ALT "\xee\xa6\x82"
#define ICON_KI_GAMEPAD_TILT_LEFT "\xee\xa6\x83"
#define ICON_KI_GAMEPAD_TILT_RIGHT "\xee\xa6\x84"
#define ICON_KI_PLAYER_ONE "\xee\xa6\x85"
#define ICON_KI_PLAYER_TWO "\xee\xa6\x86"
#define ICON_KI_PLAYER_THREE "\xee\xa6\x87"
#define ICON_KI_PLAYER_FOUR "\xee\xa6\x88"
#define ICON_KI_JOYSTICK "\xee\xa6\x89"
#define ICON_KI_JOYSTICK_ALT "\xee\xa6\x8a"
#define ICON_KI_JOYSTICK_LEFT "\xee\xa6\x8b"
#define ICON_KI_JOYSTICK_RIGHT "\xee\xa6\x8c"
#define ICON_KI_MOUSE_ALT "\xee\xa6\x8d"
#define ICON_KI_MOUSE "\xee\xa6\x8e"
#define ICON_KI_MOUSE_LEFT_BUTTON "\xee\xa6\x8f"
#define ICON_KI_MOUSE_RIGHT_BUTTON "\xee\xa6\x90"
#define ICON_KI_BUTTON_ONE "\xee\xa6\x91"
#define ICON_KI_BUTTON_TWO "\xee\xa6\x92"
#define ICON_KI_BUTTON_THREE "\xee\xa6\x93"
#define ICON_KI_BUTTON_A "\xee\xa6\x94"
#define ICON_KI_BUTTON_X "\xee\xa6\x95"
#define ICON_KI_BUTON_Y "\xee\xa6\x96"
#define ICON_KI_BUTTON_TIMES "\xee\xa6\x97"
#define ICON_KI_BUTTON_SQUARE "\xee\xa6\x98"
#define ICON_KI_BUTTON_CIRCLE "\xee\xa6\x99"
#define ICON_KI_BUTTON_TRIANGLE "\xee\xa6\x9a"
#define ICON_KI_BUTTON_LEFT "\xee\xa6\x9b"
#define ICON_KI_BUTTON_L "\xee\xa6\x9c"
#define ICON_KI_BUTTON_L1 "\xee\xa6\x9d"
#define ICON_KI_BUTTON_L2 "\xee\xa6\x9e"
#define ICON_KI_BUTTON_LB "\xee\xa6\x9f"
#define ICON_KI_BUTTON_LT "\xee\xa6\xa0"
#define ICON_KI_BUTTON_RT "\xee\xa6\xa1"
#define ICON_KI_BUTTON_RB "\xee\xa6\xa2"
#define ICON_KI_BUTTON_R2 "\xee\xa6\xa3"
#define ICON_KI_BUTTON_R1 "\xee\xa6\xa4"
#define ICON_KI_BUTTON_R "\xee\xa6\xa5"
#define ICON_KI_BUTTON_RIGHT "\xee\xa6\xa6"
#define ICON_KI_BUTTON_EMPTY "\xee\xa6\xa7"
#define ICON_KI_BUTTON_START "\xee\xa6\xa8"
#define ICON_KI_BUTTON_SELECT "\xee\xa6\xa9"
#define ICON_KI_DPAD "\xee\xa6\xaa"
#define ICON_KI_DPAD_ALT "\xee\xa6\xab"
#define ICON_KI_DPAD_TOP "\xee\xa6\xac"
#define ICON_KI_DPAD_RIGHT "\xee\xa6\xad"
#define ICON_KI_DPAD_BOTTOM "\xee\xa6\xae"
#define ICON_KI_DPAD_LEFT "\xee\xa6\xaf"
#define ICON_KI_KEY_LARGE "\xee\xa6\xb0"
#define ICON_KI_KEY_LARGE_3D "\xee\xa6\xb1"
#define ICON_KI_KEY_SMALL "\xee\xa6\xb2"
#define ICON_KI_KEY_SMALL_3D "\xee\xa6\xb3"
#define ICON_KI_STICK_LEFT_TOP "\xee\xa6\xb4"
#define ICON_KI_STICK_LEFT_SIDE "\xee\xa6\xb5"
#define ICON_KI_STICK_RIGHT_SIDE "\xee\xa6\xb6"
#define ICON_KI_STICK_RIGHT_TOP "\xee\xa6\xb7"
#define ICON_KI_STICK_SIDE "\xee\xa6\xb8"
#define ICON_KI_STICK_TILT_LEFT "\xee\xa6\xb9"
#define ICON_KI_STICK_TILT_RIGHT "\xee\xa6\xba"
#define ICON_KI_MOVE_BL "\xee\xa6\xbb"
#define ICON_KI_MOVE_BR "\xee\xa6\xbc"
#define ICON_KI_MOVE_BT "\xee\xa6\xbd"
#define ICON_KI_MOVE_BT_ALT "\xee\xa6\xbe"
#define ICON_KI_MOVE_LB "\xee\xa6\xbf"
#define ICON_KI_MOVE_LR "\xee\xa7\x80"
#define ICON_KI_MOVE_LR_ALT "\xee\xa7\x81"
#define ICON_KI_MOVE_LT "\xee\xa7\x82"
#define ICON_KI_MOVE_RB "\xee\xa7\x83"
#define ICON_KI_MOVE_RL "\xee\xa7\x84"
#define ICON_KI_MOVE_RL_ALT "\xee\xa7\x85"
#define ICON_KI_MOVE_RT "\xee\xa7\x86"
#define ICON_KI_MOVE_TB "\xee\xa7\x87"
#define ICON_KI_MOVE_TB_ALT "\xee\xa7\x88"
#define ICON_KI_MOVE_TL "\xee\xa7\x89"
#define ICON_KI_MOVE_TR "\xee\xa7\x8a"
#define ICON_KI_STICK_MOVE_BL "\xee\xa7\x8b"
#define ICON_KI_STICK_MOVE_BR "\xee\xa7\x8c"
#define ICON_KI_STICK_MOVE_BT "\xee\xa7\x8d"
#define ICON_KI_STICK_MOVE_BT_ALT "\xee\xa7\x8e"
#define ICON_KI_STICK_MOVE_LB "\xee\xa7\x8f"
#define ICON_KI_STICK_MOVE_LR "\xee\xa7\x90"
#define ICON_KI_STICK_MOVE_LR_ALT "\xee\xa7\x91"
#define ICON_KI_STICK_MOVE_LT "\xee\xa7\x92"
#define ICON_KI_STICK_MOVE_RB "\xee\xa7\x93"
#define ICON_KI_STICK_MOVE_RL "\xee\xa7\x94"
#define ICON_KI_STICK_MOVE_RL_ALT "\xee\xa7\x95"
#define ICON_KI_STICK_MOVE_RT "\xee\xa7\x96"
#define ICON_KI_STICK_MOVE_TB "\xee\xa7\x97"
#define ICON_KI_STICK_MOVE_TB_ALT "\xee\xa7\x98"
#define ICON_KI_STICK_MOVE_TL "\xee\xa7\x99"
#define ICON_KI_STICK_MOVE_TR "\xee\xa7\x9a"
#define ICON_KI_GITHUB "\xee\xa7\x9b"
#define ICON_KI_GITHUB_ALT "\xee\xa7\x9c"
#define ICON_KI_TWITTER "\xee\xa7\x9d"
#define ICON_KI_FACEBOOK "\xee\xa7\x9e"
#define ICON_KI_GOOGLE_PLUS "\xee\xa7\x9f"
#define ICON_KI_YOUTUBE "\xee\xa7\xa2"
#define ICON_KI_WE_HEART "\xee\xa7\xa3"
#define ICON_KI_WOLFCMS "\xee\xa7\xa0"
#define ICON_KI_WOLFCMS_ALT "\xee\xa7\xa1"
