// Generated by https://github.com/juliettef/IconFontCppHeaders script GenerateIconFontCppHeaders.py for language C#
// from https://raw.githubusercontent.com/nicodinh/kenney-icon-font/master/css/kenney-icons.css
// for use with https://github.com/nicodinh/kenney-icon-font/blob/master/fonts/kenney-icon-font.ttf
namespace IconFonts
{
    public class Kenney
    {
        public const string FontIconFileName = "kenney-icon-font.ttf";

        public const int IconMin = 0xe900;
        public const int IconMax = 0xe9e3;
        public const string Home = "\ue900";
        public const string Adjust = "\ue901";
        public const string Wrench = "\ue902";
        public const string Cog = "\ue903";
        public const string Off = "\ue904";
        public const string Expand = "\ue905";
        public const string Reduce = "\ue906";
        public const string Movie = "\ue907";
        public const string Flap = "\ue908";
        public const string ShoppingCart = "\ue909";
        public const string ShoppingCase = "\ue90a";
        public const string External = "\ue90b";
        public const string Network = "\ue90c";
        public const string Check = "\ue90d";
        public const string Times = "\ue90e";
        public const string TimesCircle = "\ue90f";
        public const string Plus = "\ue910";
        public const string PlusCircle = "\ue911";
        public const string Minus = "\ue912";
        public const string MinusCircle = "\ue913";
        public const string Info = "\ue914";
        public const string InfoCircle = "\ue915";
        public const string Question = "\ue916";
        public const string QuestionCircle = "\ue917";
        public const string Exlamation = "\ue918";
        public const string ExclamationCircle = "\ue919";
        public const string ExclamationTriangle = "\ue91a";
        public const string PaintBrush = "\ue91b";
        public const string Pencil = "\ue91c";
        public const string Checkbox = "\ue91d";
        public const string CheckboxChecked = "\ue91e";
        public const string Radio = "\ue91f";
        public const string RadioChecked = "\ue920";
        public const string SortVertical = "\ue921";
        public const string SortHorizontal = "\ue922";
        public const string Grid = "\ue923";
        public const string List = "\ue924";
        public const string Rows = "\ue925";
        public const string Cells = "\ue926";
        public const string SignalLow = "\ue927";
        public const string SignalMedium = "\ue928";
        public const string SignalHigh = "\ue929";
        public const string Trash = "\ue92a";
        public const string TrashAlt = "\ue92b";
        public const string ReloadInverse = "\ue92c";
        public const string Reload = "\ue92d";
        public const string Top = "\ue92e";
        public const string Bottom = "\ue92f";
        public const string Upload = "\ue930";
        public const string Download = "\ue931";
        public const string Cloud = "\ue932";
        public const string CloudUpload = "\ue933";
        public const string CloudDownload = "\ue934";
        public const string Search = "\ue935";
        public const string SearchPlus = "\ue936";
        public const string SearchMinus = "\ue937";
        public const string SearchEqual = "\ue938";
        public const string Lock = "\ue939";
        public const string Unlock = "\ue93a";
        public const string User = "\ue93b";
        public const string Users = "\ue93c";
        public const string UsersAlt = "\ue93d";
        public const string SignIn = "\ue93e";
        public const string SignInInverse = "\ue93f";
        public const string SignOut = "\ue940";
        public const string SignOutInverse = "\ue941";
        public const string ArrowTop = "\ue942";
        public const string ArrowRight = "\ue943";
        public const string ArrowBottom = "\ue944";
        public const string ArrowLeft = "\ue945";
        public const string ArrowTopLeft = "\ue946";
        public const string ArrowTopRight = "\ue947";
        public const string ArrowBottomRight = "\ue948";
        public const string ArrowBottomLeft = "\ue949";
        public const string CaretTop = "\ue94a";
        public const string CaretRight = "\ue94b";
        public const string CaretBottom = "\ue94c";
        public const string CaretLeft = "\ue94d";
        public const string NextAlt = "\ue94e";
        public const string Next = "\ue94f";
        public const string Previous = "\ue950";
        public const string PreviousAlt = "\ue951";
        public const string Fill = "\ue952";
        public const string Eraser = "\ue953";
        public const string Save = "\ue954";
        public const string StepBackward = "\ue955";
        public const string Backward = "\ue956";
        public const string Pause = "\ue957";
        public const string Forward = "\ue958";
        public const string StepForward = "\ue959";
        public const string Stop = "\ue95a";
        public const string Rec = "\ue95b";
        public const string Cursor = "\ue95c";
        public const string Pointer = "\ue95d";
        public const string Exit = "\ue95e";
        public const string Figure = "\ue95f";
        public const string Car = "\ue960";
        public const string Coin = "\ue961";
        public const string Key = "\ue962";
        public const string Cub = "\ue963";
        public const string Diamond = "\ue964";
        public const string Badge = "\ue965";
        public const string BadgeAlt = "\ue966";
        public const string Podium = "\ue967";
        public const string PodiumAlt = "\ue968";
        public const string Flag = "\ue969";
        public const string Fist = "\ue96a";
        public const string FistCircle = "\ue96b";
        public const string Heart = "\ue96c";
        public const string HeartHalf = "\ue96d";
        public const string HeartHalfO = "\ue96e";
        public const string HeartO = "\ue96f";
        public const string Star = "\ue970";
        public const string StarHalf = "\ue971";
        public const string StarHalfO = "\ue972";
        public const string StarO = "\ue973";
        public const string ButtonB = "\ue974";
        public const string MusicOn = "\ue975";
        public const string MusicOff = "\ue976";
        public const string SoundOn = "\ue977";
        public const string SoundOff = "\ue978";
        public const string SoundOffAlt = "\ue979";
        public const string Robot = "\ue97a";
        public const string Computer = "\ue97b";
        public const string Tablet = "\ue97c";
        public const string Smartphone = "\ue97d";
        public const string Device = "\ue97e";
        public const string DeviceTiltLeft = "\ue97f";
        public const string DeviceTiltRight = "\ue980";
        public const string Gamepad = "\ue981";
        public const string GamepadAlt = "\ue982";
        public const string GamepadTiltLeft = "\ue983";
        public const string GamepadTiltRight = "\ue984";
        public const string PlayerOne = "\ue985";
        public const string PlayerTwo = "\ue986";
        public const string PlayerThree = "\ue987";
        public const string PlayerFour = "\ue988";
        public const string Joystick = "\ue989";
        public const string JoystickAlt = "\ue98a";
        public const string JoystickLeft = "\ue98b";
        public const string JoystickRight = "\ue98c";
        public const string MouseAlt = "\ue98d";
        public const string Mouse = "\ue98e";
        public const string MouseLeftButton = "\ue98f";
        public const string MouseRightButton = "\ue990";
        public const string ButtonOne = "\ue991";
        public const string ButtonTwo = "\ue992";
        public const string ButtonThree = "\ue993";
        public const string ButtonA = "\ue994";
        public const string ButtonX = "\ue995";
        public const string ButonY = "\ue996";
        public const string ButtonTimes = "\ue997";
        public const string ButtonSquare = "\ue998";
        public const string ButtonCircle = "\ue999";
        public const string ButtonTriangle = "\ue99a";
        public const string ButtonLeft = "\ue99b";
        public const string ButtonL = "\ue99c";
        public const string ButtonL1 = "\ue99d";
        public const string ButtonL2 = "\ue99e";
        public const string ButtonLb = "\ue99f";
        public const string ButtonLt = "\ue9a0";
        public const string ButtonRt = "\ue9a1";
        public const string ButtonRb = "\ue9a2";
        public const string ButtonR2 = "\ue9a3";
        public const string ButtonR1 = "\ue9a4";
        public const string ButtonR = "\ue9a5";
        public const string ButtonRight = "\ue9a6";
        public const string ButtonEmpty = "\ue9a7";
        public const string ButtonStart = "\ue9a8";
        public const string ButtonSelect = "\ue9a9";
        public const string Dpad = "\ue9aa";
        public const string DpadAlt = "\ue9ab";
        public const string DpadTop = "\ue9ac";
        public const string DpadRight = "\ue9ad";
        public const string DpadBottom = "\ue9ae";
        public const string DpadLeft = "\ue9af";
        public const string KeyLarge = "\ue9b0";
        public const string KeyLarge3d = "\ue9b1";
        public const string KeySmall = "\ue9b2";
        public const string KeySmall3d = "\ue9b3";
        public const string StickLeftTop = "\ue9b4";
        public const string StickLeftSide = "\ue9b5";
        public const string StickRightSide = "\ue9b6";
        public const string StickRightTop = "\ue9b7";
        public const string StickSide = "\ue9b8";
        public const string StickTiltLeft = "\ue9b9";
        public const string StickTiltRight = "\ue9ba";
        public const string MoveBl = "\ue9bb";
        public const string MoveBr = "\ue9bc";
        public const string MoveBt = "\ue9bd";
        public const string MoveBtAlt = "\ue9be";
        public const string MoveLb = "\ue9bf";
        public const string MoveLr = "\ue9c0";
        public const string MoveLrAlt = "\ue9c1";
        public const string MoveLt = "\ue9c2";
        public const string MoveRb = "\ue9c3";
        public const string MoveRl = "\ue9c4";
        public const string MoveRlAlt = "\ue9c5";
        public const string MoveRt = "\ue9c6";
        public const string MoveTb = "\ue9c7";
        public const string MoveTbAlt = "\ue9c8";
        public const string MoveTl = "\ue9c9";
        public const string MoveTr = "\ue9ca";
        public const string StickMoveBl = "\ue9cb";
        public const string StickMoveBr = "\ue9cc";
        public const string StickMoveBt = "\ue9cd";
        public const string StickMoveBtAlt = "\ue9ce";
        public const string StickMoveLb = "\ue9cf";
        public const string StickMoveLr = "\ue9d0";
        public const string StickMoveLrAlt = "\ue9d1";
        public const string StickMoveLt = "\ue9d2";
        public const string StickMoveRb = "\ue9d3";
        public const string StickMoveRl = "\ue9d4";
        public const string StickMoveRlAlt = "\ue9d5";
        public const string StickMoveRt = "\ue9d6";
        public const string StickMoveTb = "\ue9d7";
        public const string StickMoveTbAlt = "\ue9d8";
        public const string StickMoveTl = "\ue9d9";
        public const string StickMoveTr = "\ue9da";
        public const string Github = "\ue9db";
        public const string GithubAlt = "\ue9dc";
        public const string Twitter = "\ue9dd";
        public const string Facebook = "\ue9de";
        public const string GooglePlus = "\ue9df";
        public const string Youtube = "\ue9e2";
        public const string WeHeart = "\ue9e3";
        public const string Wolfcms = "\ue9e0";
        public const string WolfcmsAlt = "\ue9e1";
    }
}
