// Generated by https://github.com/juliettef/IconFontCppHeaders script GenerateIconFontCppHeaders.py for language C89
// from https://raw.githubusercontent.com/FortAwesome/Font-Awesome/master/metadata/icons.yml
// for use with https://github.com/FortAwesome/Font-Awesome/blob/master/webfonts/fa-brands-400.ttf
#pragma once

#define FONT_ICON_FILE_NAME_FAB "fa-brands-400.ttf"

#define ICON_MIN_FAB 0xe007
#define ICON_MAX_FAB 0xf8e8
#define ICON_FA_500PX "\xef\x89\xae"
#define ICON_FA_ACCESSIBLE_ICON "\xef\x8d\xa8"
#define ICON_FA_ACCUSOFT "\xef\x8d\xa9"
#define ICON_FA_ACQUISITIONS_INCORPORATED "\xef\x9a\xaf"
#define ICON_FA_ADN "\xef\x85\xb0"
#define ICON_FA_ADOBE "\xef\x9d\xb8"
#define ICON_FA_ADVERSAL "\xef\x8d\xaa"
#define ICON_FA_AFFILIATETHEME "\xef\x8d\xab"
#define ICON_FA_AIRBNB "\xef\xa0\xb4"
#define ICON_FA_ALGOLIA "\xef\x8d\xac"
#define ICON_FA_ALIPAY "\xef\x99\x82"
#define ICON_FA_AMAZON "\xef\x89\xb0"
#define ICON_FA_AMAZON_PAY "\xef\x90\xac"
#define ICON_FA_AMILIA "\xef\x8d\xad"
#define ICON_FA_ANDROID "\xef\x85\xbb"
#define ICON_FA_ANGELLIST "\xef\x88\x89"
#define ICON_FA_ANGRYCREATIVE "\xef\x8d\xae"
#define ICON_FA_ANGULAR "\xef\x90\xa0"
#define ICON_FA_APP_STORE "\xef\x8d\xaf"
#define ICON_FA_APP_STORE_IOS "\xef\x8d\xb0"
#define ICON_FA_APPER "\xef\x8d\xb1"
#define ICON_FA_APPLE "\xef\x85\xb9"
#define ICON_FA_APPLE_PAY "\xef\x90\x95"
#define ICON_FA_ARTSTATION "\xef\x9d\xba"
#define ICON_FA_ASYMMETRIK "\xef\x8d\xb2"
#define ICON_FA_ATLASSIAN "\xef\x9d\xbb"
#define ICON_FA_AUDIBLE "\xef\x8d\xb3"
#define ICON_FA_AUTOPREFIXER "\xef\x90\x9c"
#define ICON_FA_AVIANEX "\xef\x8d\xb4"
#define ICON_FA_AVIATO "\xef\x90\xa1"
#define ICON_FA_AWS "\xef\x8d\xb5"
#define ICON_FA_BANDCAMP "\xef\x8b\x95"
#define ICON_FA_BATTLE_NET "\xef\xa0\xb5"
#define ICON_FA_BEHANCE "\xef\x86\xb4"
#define ICON_FA_BEHANCE_SQUARE "\xef\x86\xb5"
#define ICON_FA_BIMOBJECT "\xef\x8d\xb8"
#define ICON_FA_BITBUCKET "\xef\x85\xb1"
#define ICON_FA_BITCOIN "\xef\x8d\xb9"
#define ICON_FA_BITY "\xef\x8d\xba"
#define ICON_FA_BLACK_TIE "\xef\x89\xbe"
#define ICON_FA_BLACKBERRY "\xef\x8d\xbb"
#define ICON_FA_BLOGGER "\xef\x8d\xbc"
#define ICON_FA_BLOGGER_B "\xef\x8d\xbd"
#define ICON_FA_BLUETOOTH "\xef\x8a\x93"
#define ICON_FA_BLUETOOTH_B "\xef\x8a\x94"
#define ICON_FA_BOOTSTRAP "\xef\xa0\xb6"
#define ICON_FA_BTC "\xef\x85\x9a"
#define ICON_FA_BUFFER "\xef\xa0\xb7"
#define ICON_FA_BUROMOBELEXPERTE "\xef\x8d\xbf"
#define ICON_FA_BUY_N_LARGE "\xef\xa2\xa6"
#define ICON_FA_BUYSELLADS "\xef\x88\x8d"
#define ICON_FA_CANADIAN_MAPLE_LEAF "\xef\x9e\x85"
#define ICON_FA_CC_AMAZON_PAY "\xef\x90\xad"
#define ICON_FA_CC_AMEX "\xef\x87\xb3"
#define ICON_FA_CC_APPLE_PAY "\xef\x90\x96"
#define ICON_FA_CC_DINERS_CLUB "\xef\x89\x8c"
#define ICON_FA_CC_DISCOVER "\xef\x87\xb2"
#define ICON_FA_CC_JCB "\xef\x89\x8b"
#define ICON_FA_CC_MASTERCARD "\xef\x87\xb1"
#define ICON_FA_CC_PAYPAL "\xef\x87\xb4"
#define ICON_FA_CC_STRIPE "\xef\x87\xb5"
#define ICON_FA_CC_VISA "\xef\x87\xb0"
#define ICON_FA_CENTERCODE "\xef\x8e\x80"
#define ICON_FA_CENTOS "\xef\x9e\x89"
#define ICON_FA_CHROME "\xef\x89\xa8"
#define ICON_FA_CHROMECAST "\xef\xa0\xb8"
#define ICON_FA_CLOUDSCALE "\xef\x8e\x83"
#define ICON_FA_CLOUDSMITH "\xef\x8e\x84"
#define ICON_FA_CLOUDVERSIFY "\xef\x8e\x85"
#define ICON_FA_CODEPEN "\xef\x87\x8b"
#define ICON_FA_CODIEPIE "\xef\x8a\x84"
#define ICON_FA_CONFLUENCE "\xef\x9e\x8d"
#define ICON_FA_CONNECTDEVELOP "\xef\x88\x8e"
#define ICON_FA_CONTAO "\xef\x89\xad"
#define ICON_FA_COTTON_BUREAU "\xef\xa2\x9e"
#define ICON_FA_CPANEL "\xef\x8e\x88"
#define ICON_FA_CREATIVE_COMMONS "\xef\x89\x9e"
#define ICON_FA_CREATIVE_COMMONS_BY "\xef\x93\xa7"
#define ICON_FA_CREATIVE_COMMONS_NC "\xef\x93\xa8"
#define ICON_FA_CREATIVE_COMMONS_NC_EU "\xef\x93\xa9"
#define ICON_FA_CREATIVE_COMMONS_NC_JP "\xef\x93\xaa"
#define ICON_FA_CREATIVE_COMMONS_ND "\xef\x93\xab"
#define ICON_FA_CREATIVE_COMMONS_PD "\xef\x93\xac"
#define ICON_FA_CREATIVE_COMMONS_PD_ALT "\xef\x93\xad"
#define ICON_FA_CREATIVE_COMMONS_REMIX "\xef\x93\xae"
#define ICON_FA_CREATIVE_COMMONS_SA "\xef\x93\xaf"
#define ICON_FA_CREATIVE_COMMONS_SAMPLING "\xef\x93\xb0"
#define ICON_FA_CREATIVE_COMMONS_SAMPLING_PLUS "\xef\x93\xb1"
#define ICON_FA_CREATIVE_COMMONS_SHARE "\xef\x93\xb2"
#define ICON_FA_CREATIVE_COMMONS_ZERO "\xef\x93\xb3"
#define ICON_FA_CRITICAL_ROLE "\xef\x9b\x89"
#define ICON_FA_CSS3 "\xef\x84\xbc"
#define ICON_FA_CSS3_ALT "\xef\x8e\x8b"
#define ICON_FA_CUTTLEFISH "\xef\x8e\x8c"
#define ICON_FA_D_AND_D "\xef\x8e\x8d"
#define ICON_FA_D_AND_D_BEYOND "\xef\x9b\x8a"
#define ICON_FA_DAILYMOTION "\xee\x81\x92"
#define ICON_FA_DASHCUBE "\xef\x88\x90"
#define ICON_FA_DEEZER "\xee\x81\xb7"
#define ICON_FA_DELICIOUS "\xef\x86\xa5"
#define ICON_FA_DEPLOYDOG "\xef\x8e\x8e"
#define ICON_FA_DESKPRO "\xef\x8e\x8f"
#define ICON_FA_DEV "\xef\x9b\x8c"
#define ICON_FA_DEVIANTART "\xef\x86\xbd"
#define ICON_FA_DHL "\xef\x9e\x90"
#define ICON_FA_DIASPORA "\xef\x9e\x91"
#define ICON_FA_DIGG "\xef\x86\xa6"
#define ICON_FA_DIGITAL_OCEAN "\xef\x8e\x91"
#define ICON_FA_DISCORD "\xef\x8e\x92"
#define ICON_FA_DISCOURSE "\xef\x8e\x93"
#define ICON_FA_DOCHUB "\xef\x8e\x94"
#define ICON_FA_DOCKER "\xef\x8e\x95"
#define ICON_FA_DRAFT2DIGITAL "\xef\x8e\x96"
#define ICON_FA_DRIBBBLE "\xef\x85\xbd"
#define ICON_FA_DRIBBBLE_SQUARE "\xef\x8e\x97"
#define ICON_FA_DROPBOX "\xef\x85\xab"
#define ICON_FA_DRUPAL "\xef\x86\xa9"
#define ICON_FA_DYALOG "\xef\x8e\x99"
#define ICON_FA_EARLYBIRDS "\xef\x8e\x9a"
#define ICON_FA_EBAY "\xef\x93\xb4"
#define ICON_FA_EDGE "\xef\x8a\x82"
#define ICON_FA_EDGE_LEGACY "\xee\x81\xb8"
#define ICON_FA_ELEMENTOR "\xef\x90\xb0"
#define ICON_FA_ELLO "\xef\x97\xb1"
#define ICON_FA_EMBER "\xef\x90\xa3"
#define ICON_FA_EMPIRE "\xef\x87\x91"
#define ICON_FA_ENVIRA "\xef\x8a\x99"
#define ICON_FA_ERLANG "\xef\x8e\x9d"
#define ICON_FA_ETHEREUM "\xef\x90\xae"
#define ICON_FA_ETSY "\xef\x8b\x97"
#define ICON_FA_EVERNOTE "\xef\xa0\xb9"
#define ICON_FA_EXPEDITEDSSL "\xef\x88\xbe"
#define ICON_FA_FACEBOOK "\xef\x82\x9a"
#define ICON_FA_FACEBOOK_F "\xef\x8e\x9e"
#define ICON_FA_FACEBOOK_MESSENGER "\xef\x8e\x9f"
#define ICON_FA_FACEBOOK_SQUARE "\xef\x82\x82"
#define ICON_FA_FANTASY_FLIGHT_GAMES "\xef\x9b\x9c"
#define ICON_FA_FEDEX "\xef\x9e\x97"
#define ICON_FA_FEDORA "\xef\x9e\x98"
#define ICON_FA_FIGMA "\xef\x9e\x99"
#define ICON_FA_FIREFOX "\xef\x89\xa9"
#define ICON_FA_FIREFOX_BROWSER "\xee\x80\x87"
#define ICON_FA_FIRST_ORDER "\xef\x8a\xb0"
#define ICON_FA_FIRST_ORDER_ALT "\xef\x94\x8a"
#define ICON_FA_FIRSTDRAFT "\xef\x8e\xa1"
#define ICON_FA_FLICKR "\xef\x85\xae"
#define ICON_FA_FLIPBOARD "\xef\x91\x8d"
#define ICON_FA_FLY "\xef\x90\x97"
#define ICON_FA_FONT_AWESOME "\xef\x8a\xb4"
#define ICON_FA_FONT_AWESOME_ALT "\xef\x8d\x9c"
#define ICON_FA_FONT_AWESOME_FLAG "\xef\x90\xa5"
#define ICON_FA_FONT_AWESOME_LOGO_FULL "\xef\x93\xa6"
#define ICON_FA_FONTICONS "\xef\x8a\x80"
#define ICON_FA_FONTICONS_FI "\xef\x8e\xa2"
#define ICON_FA_FORT_AWESOME "\xef\x8a\x86"
#define ICON_FA_FORT_AWESOME_ALT "\xef\x8e\xa3"
#define ICON_FA_FORUMBEE "\xef\x88\x91"
#define ICON_FA_FOURSQUARE "\xef\x86\x80"
#define ICON_FA_FREE_CODE_CAMP "\xef\x8b\x85"
#define ICON_FA_FREEBSD "\xef\x8e\xa4"
#define ICON_FA_FULCRUM "\xef\x94\x8b"
#define ICON_FA_GALACTIC_REPUBLIC "\xef\x94\x8c"
#define ICON_FA_GALACTIC_SENATE "\xef\x94\x8d"
#define ICON_FA_GET_POCKET "\xef\x89\xa5"
#define ICON_FA_GG "\xef\x89\xa0"
#define ICON_FA_GG_CIRCLE "\xef\x89\xa1"
#define ICON_FA_GIT "\xef\x87\x93"
#define ICON_FA_GIT_ALT "\xef\xa1\x81"
#define ICON_FA_GIT_SQUARE "\xef\x87\x92"
#define ICON_FA_GITHUB "\xef\x82\x9b"
#define ICON_FA_GITHUB_ALT "\xef\x84\x93"
#define ICON_FA_GITHUB_SQUARE "\xef\x82\x92"
#define ICON_FA_GITKRAKEN "\xef\x8e\xa6"
#define ICON_FA_GITLAB "\xef\x8a\x96"
#define ICON_FA_GITTER "\xef\x90\xa6"
#define ICON_FA_GLIDE "\xef\x8a\xa5"
#define ICON_FA_GLIDE_G "\xef\x8a\xa6"
#define ICON_FA_GOFORE "\xef\x8e\xa7"
#define ICON_FA_GOODREADS "\xef\x8e\xa8"
#define ICON_FA_GOODREADS_G "\xef\x8e\xa9"
#define ICON_FA_GOOGLE "\xef\x86\xa0"
#define ICON_FA_GOOGLE_DRIVE "\xef\x8e\xaa"
#define ICON_FA_GOOGLE_PAY "\xee\x81\xb9"
#define ICON_FA_GOOGLE_PLAY "\xef\x8e\xab"
#define ICON_FA_GOOGLE_PLUS "\xef\x8a\xb3"
#define ICON_FA_GOOGLE_PLUS_G "\xef\x83\x95"
#define ICON_FA_GOOGLE_PLUS_SQUARE "\xef\x83\x94"
#define ICON_FA_GOOGLE_WALLET "\xef\x87\xae"
#define ICON_FA_GRATIPAY "\xef\x86\x84"
#define ICON_FA_GRAV "\xef\x8b\x96"
#define ICON_FA_GRIPFIRE "\xef\x8e\xac"
#define ICON_FA_GRUNT "\xef\x8e\xad"
#define ICON_FA_GULP "\xef\x8e\xae"
#define ICON_FA_HACKER_NEWS "\xef\x87\x94"
#define ICON_FA_HACKER_NEWS_SQUARE "\xef\x8e\xaf"
#define ICON_FA_HACKERRANK "\xef\x97\xb7"
#define ICON_FA_HIPS "\xef\x91\x92"
#define ICON_FA_HIRE_A_HELPER "\xef\x8e\xb0"
#define ICON_FA_HOOLI "\xef\x90\xa7"
#define ICON_FA_HORNBILL "\xef\x96\x92"
#define ICON_FA_HOTJAR "\xef\x8e\xb1"
#define ICON_FA_HOUZZ "\xef\x89\xbc"
#define ICON_FA_HTML5 "\xef\x84\xbb"
#define ICON_FA_HUBSPOT "\xef\x8e\xb2"
#define ICON_FA_IDEAL "\xee\x80\x93"
#define ICON_FA_IMDB "\xef\x8b\x98"
#define ICON_FA_INSTAGRAM "\xef\x85\xad"
#define ICON_FA_INSTAGRAM_SQUARE "\xee\x81\x95"
#define ICON_FA_INTERCOM "\xef\x9e\xaf"
#define ICON_FA_INTERNET_EXPLORER "\xef\x89\xab"
#define ICON_FA_INVISION "\xef\x9e\xb0"
#define ICON_FA_IOXHOST "\xef\x88\x88"
#define ICON_FA_ITCH_IO "\xef\xa0\xba"
#define ICON_FA_ITUNES "\xef\x8e\xb4"
#define ICON_FA_ITUNES_NOTE "\xef\x8e\xb5"
#define ICON_FA_JAVA "\xef\x93\xa4"
#define ICON_FA_JEDI_ORDER "\xef\x94\x8e"
#define ICON_FA_JENKINS "\xef\x8e\xb6"
#define ICON_FA_JIRA "\xef\x9e\xb1"
#define ICON_FA_JOGET "\xef\x8e\xb7"
#define ICON_FA_JOOMLA "\xef\x86\xaa"
#define ICON_FA_JS "\xef\x8e\xb8"
#define ICON_FA_JS_SQUARE "\xef\x8e\xb9"
#define ICON_FA_JSFIDDLE "\xef\x87\x8c"
#define ICON_FA_KAGGLE "\xef\x97\xba"
#define ICON_FA_KEYBASE "\xef\x93\xb5"
#define ICON_FA_KEYCDN "\xef\x8e\xba"
#define ICON_FA_KICKSTARTER "\xef\x8e\xbb"
#define ICON_FA_KICKSTARTER_K "\xef\x8e\xbc"
#define ICON_FA_KORVUE "\xef\x90\xaf"
#define ICON_FA_LARAVEL "\xef\x8e\xbd"
#define ICON_FA_LASTFM "\xef\x88\x82"
#define ICON_FA_LASTFM_SQUARE "\xef\x88\x83"
#define ICON_FA_LEANPUB "\xef\x88\x92"
#define ICON_FA_LESS "\xef\x90\x9d"
#define ICON_FA_LINE "\xef\x8f\x80"
#define ICON_FA_LINKEDIN "\xef\x82\x8c"
#define ICON_FA_LINKEDIN_IN "\xef\x83\xa1"
#define ICON_FA_LINODE "\xef\x8a\xb8"
#define ICON_FA_LINUX "\xef\x85\xbc"
#define ICON_FA_LYFT "\xef\x8f\x83"
#define ICON_FA_MAGENTO "\xef\x8f\x84"
#define ICON_FA_MAILCHIMP "\xef\x96\x9e"
#define ICON_FA_MANDALORIAN "\xef\x94\x8f"
#define ICON_FA_MARKDOWN "\xef\x98\x8f"
#define ICON_FA_MASTODON "\xef\x93\xb6"
#define ICON_FA_MAXCDN "\xef\x84\xb6"
#define ICON_FA_MDB "\xef\xa3\x8a"
#define ICON_FA_MEDAPPS "\xef\x8f\x86"
#define ICON_FA_MEDIUM "\xef\x88\xba"
#define ICON_FA_MEDIUM_M "\xef\x8f\x87"
#define ICON_FA_MEDRT "\xef\x8f\x88"
#define ICON_FA_MEETUP "\xef\x8b\xa0"
#define ICON_FA_MEGAPORT "\xef\x96\xa3"
#define ICON_FA_MENDELEY "\xef\x9e\xb3"
#define ICON_FA_MICROBLOG "\xee\x80\x9a"
#define ICON_FA_MICROSOFT "\xef\x8f\x8a"
#define ICON_FA_MIX "\xef\x8f\x8b"
#define ICON_FA_MIXCLOUD "\xef\x8a\x89"
#define ICON_FA_MIXER "\xee\x81\x96"
#define ICON_FA_MIZUNI "\xef\x8f\x8c"
#define ICON_FA_MODX "\xef\x8a\x85"
#define ICON_FA_MONERO "\xef\x8f\x90"
#define ICON_FA_NAPSTER "\xef\x8f\x92"
#define ICON_FA_NEOS "\xef\x98\x92"
#define ICON_FA_NIMBLR "\xef\x96\xa8"
#define ICON_FA_NODE "\xef\x90\x99"
#define ICON_FA_NODE_JS "\xef\x8f\x93"
#define ICON_FA_NPM "\xef\x8f\x94"
#define ICON_FA_NS8 "\xef\x8f\x95"
#define ICON_FA_NUTRITIONIX "\xef\x8f\x96"
#define ICON_FA_ODNOKLASSNIKI "\xef\x89\xa3"
#define ICON_FA_ODNOKLASSNIKI_SQUARE "\xef\x89\xa4"
#define ICON_FA_OLD_REPUBLIC "\xef\x94\x90"
#define ICON_FA_OPENCART "\xef\x88\xbd"
#define ICON_FA_OPENID "\xef\x86\x9b"
#define ICON_FA_OPERA "\xef\x89\xaa"
#define ICON_FA_OPTIN_MONSTER "\xef\x88\xbc"
#define ICON_FA_ORCID "\xef\xa3\x92"
#define ICON_FA_OSI "\xef\x90\x9a"
#define ICON_FA_PAGE4 "\xef\x8f\x97"
#define ICON_FA_PAGELINES "\xef\x86\x8c"
#define ICON_FA_PALFED "\xef\x8f\x98"
#define ICON_FA_PATREON "\xef\x8f\x99"
#define ICON_FA_PAYPAL "\xef\x87\xad"
#define ICON_FA_PENNY_ARCADE "\xef\x9c\x84"
#define ICON_FA_PERISCOPE "\xef\x8f\x9a"
#define ICON_FA_PHABRICATOR "\xef\x8f\x9b"
#define ICON_FA_PHOENIX_FRAMEWORK "\xef\x8f\x9c"
#define ICON_FA_PHOENIX_SQUADRON "\xef\x94\x91"
#define ICON_FA_PHP "\xef\x91\x97"
#define ICON_FA_PIED_PIPER "\xef\x8a\xae"
#define ICON_FA_PIED_PIPER_ALT "\xef\x86\xa8"
#define ICON_FA_PIED_PIPER_HAT "\xef\x93\xa5"
#define ICON_FA_PIED_PIPER_PP "\xef\x86\xa7"
#define ICON_FA_PIED_PIPER_SQUARE "\xee\x80\x9e"
#define ICON_FA_PINTEREST "\xef\x83\x92"
#define ICON_FA_PINTEREST_P "\xef\x88\xb1"
#define ICON_FA_PINTEREST_SQUARE "\xef\x83\x93"
#define ICON_FA_PLAYSTATION "\xef\x8f\x9f"
#define ICON_FA_PRODUCT_HUNT "\xef\x8a\x88"
#define ICON_FA_PUSHED "\xef\x8f\xa1"
#define ICON_FA_PYTHON "\xef\x8f\xa2"
#define ICON_FA_QQ "\xef\x87\x96"
#define ICON_FA_QUINSCAPE "\xef\x91\x99"
#define ICON_FA_QUORA "\xef\x8b\x84"
#define ICON_FA_R_PROJECT "\xef\x93\xb7"
#define ICON_FA_RASPBERRY_PI "\xef\x9e\xbb"
#define ICON_FA_RAVELRY "\xef\x8b\x99"
#define ICON_FA_REACT "\xef\x90\x9b"
#define ICON_FA_REACTEUROPE "\xef\x9d\x9d"
#define ICON_FA_README "\xef\x93\x95"
#define ICON_FA_REBEL "\xef\x87\x90"
#define ICON_FA_RED_RIVER "\xef\x8f\xa3"
#define ICON_FA_REDDIT "\xef\x86\xa1"
#define ICON_FA_REDDIT_ALIEN "\xef\x8a\x81"
#define ICON_FA_REDDIT_SQUARE "\xef\x86\xa2"
#define ICON_FA_REDHAT "\xef\x9e\xbc"
#define ICON_FA_RENREN "\xef\x86\x8b"
#define ICON_FA_REPLYD "\xef\x8f\xa6"
#define ICON_FA_RESEARCHGATE "\xef\x93\xb8"
#define ICON_FA_RESOLVING "\xef\x8f\xa7"
#define ICON_FA_REV "\xef\x96\xb2"
#define ICON_FA_ROCKETCHAT "\xef\x8f\xa8"
#define ICON_FA_ROCKRMS "\xef\x8f\xa9"
#define ICON_FA_RUST "\xee\x81\xba"
#define ICON_FA_SAFARI "\xef\x89\xa7"
#define ICON_FA_SALESFORCE "\xef\xa0\xbb"
#define ICON_FA_SASS "\xef\x90\x9e"
#define ICON_FA_SCHLIX "\xef\x8f\xaa"
#define ICON_FA_SCRIBD "\xef\x8a\x8a"
#define ICON_FA_SEARCHENGIN "\xef\x8f\xab"
#define ICON_FA_SELLCAST "\xef\x8b\x9a"
#define ICON_FA_SELLSY "\xef\x88\x93"
#define ICON_FA_SERVICESTACK "\xef\x8f\xac"
#define ICON_FA_SHIRTSINBULK "\xef\x88\x94"
#define ICON_FA_SHOPIFY "\xee\x81\x97"
#define ICON_FA_SHOPWARE "\xef\x96\xb5"
#define ICON_FA_SIMPLYBUILT "\xef\x88\x95"
#define ICON_FA_SISTRIX "\xef\x8f\xae"
#define ICON_FA_SITH "\xef\x94\x92"
#define ICON_FA_SKETCH "\xef\x9f\x86"
#define ICON_FA_SKYATLAS "\xef\x88\x96"
#define ICON_FA_SKYPE "\xef\x85\xbe"
#define ICON_FA_SLACK "\xef\x86\x98"
#define ICON_FA_SLACK_HASH "\xef\x8f\xaf"
#define ICON_FA_SLIDESHARE "\xef\x87\xa7"
#define ICON_FA_SNAPCHAT "\xef\x8a\xab"
#define ICON_FA_SNAPCHAT_GHOST "\xef\x8a\xac"
#define ICON_FA_SNAPCHAT_SQUARE "\xef\x8a\xad"
#define ICON_FA_SOUNDCLOUD "\xef\x86\xbe"
#define ICON_FA_SOURCETREE "\xef\x9f\x93"
#define ICON_FA_SPEAKAP "\xef\x8f\xb3"
#define ICON_FA_SPEAKER_DECK "\xef\xa0\xbc"
#define ICON_FA_SPOTIFY "\xef\x86\xbc"
#define ICON_FA_SQUARESPACE "\xef\x96\xbe"
#define ICON_FA_STACK_EXCHANGE "\xef\x86\x8d"
#define ICON_FA_STACK_OVERFLOW "\xef\x85\xac"
#define ICON_FA_STACKPATH "\xef\xa1\x82"
#define ICON_FA_STAYLINKED "\xef\x8f\xb5"
#define ICON_FA_STEAM "\xef\x86\xb6"
#define ICON_FA_STEAM_SQUARE "\xef\x86\xb7"
#define ICON_FA_STEAM_SYMBOL "\xef\x8f\xb6"
#define ICON_FA_STICKER_MULE "\xef\x8f\xb7"
#define ICON_FA_STRAVA "\xef\x90\xa8"
#define ICON_FA_STRIPE "\xef\x90\xa9"
#define ICON_FA_STRIPE_S "\xef\x90\xaa"
#define ICON_FA_STUDIOVINARI "\xef\x8f\xb8"
#define ICON_FA_STUMBLEUPON "\xef\x86\xa4"
#define ICON_FA_STUMBLEUPON_CIRCLE "\xef\x86\xa3"
#define ICON_FA_SUPERPOWERS "\xef\x8b\x9d"
#define ICON_FA_SUPPLE "\xef\x8f\xb9"
#define ICON_FA_SUSE "\xef\x9f\x96"
#define ICON_FA_SWIFT "\xef\xa3\xa1"
#define ICON_FA_SYMFONY "\xef\xa0\xbd"
#define ICON_FA_TEAMSPEAK "\xef\x93\xb9"
#define ICON_FA_TELEGRAM "\xef\x8b\x86"
#define ICON_FA_TELEGRAM_PLANE "\xef\x8f\xbe"
#define ICON_FA_TENCENT_WEIBO "\xef\x87\x95"
#define ICON_FA_THE_RED_YETI "\xef\x9a\x9d"
#define ICON_FA_THEMECO "\xef\x97\x86"
#define ICON_FA_THEMEISLE "\xef\x8a\xb2"
#define ICON_FA_THINK_PEAKS "\xef\x9c\xb1"
#define ICON_FA_TIKTOK "\xee\x81\xbb"
#define ICON_FA_TRADE_FEDERATION "\xef\x94\x93"
#define ICON_FA_TRELLO "\xef\x86\x81"
#define ICON_FA_TRIPADVISOR "\xef\x89\xa2"
#define ICON_FA_TUMBLR "\xef\x85\xb3"
#define ICON_FA_TUMBLR_SQUARE "\xef\x85\xb4"
#define ICON_FA_TWITCH "\xef\x87\xa8"
#define ICON_FA_TWITTER "\xef\x82\x99"
#define ICON_FA_TWITTER_SQUARE "\xef\x82\x81"
#define ICON_FA_TYPO3 "\xef\x90\xab"
#define ICON_FA_UBER "\xef\x90\x82"
#define ICON_FA_UBUNTU "\xef\x9f\x9f"
#define ICON_FA_UIKIT "\xef\x90\x83"
#define ICON_FA_UMBRACO "\xef\xa3\xa8"
#define ICON_FA_UNIREGISTRY "\xef\x90\x84"
#define ICON_FA_UNITY "\xee\x81\x89"
#define ICON_FA_UNSPLASH "\xee\x81\xbc"
#define ICON_FA_UNTAPPD "\xef\x90\x85"
#define ICON_FA_UPS "\xef\x9f\xa0"
#define ICON_FA_USB "\xef\x8a\x87"
#define ICON_FA_USPS "\xef\x9f\xa1"
#define ICON_FA_USSUNNAH "\xef\x90\x87"
#define ICON_FA_VAADIN "\xef\x90\x88"
#define ICON_FA_VIACOIN "\xef\x88\xb7"
#define ICON_FA_VIADEO "\xef\x8a\xa9"
#define ICON_FA_VIADEO_SQUARE "\xef\x8a\xaa"
#define ICON_FA_VIBER "\xef\x90\x89"
#define ICON_FA_VIMEO "\xef\x90\x8a"
#define ICON_FA_VIMEO_SQUARE "\xef\x86\x94"
#define ICON_FA_VIMEO_V "\xef\x89\xbd"
#define ICON_FA_VINE "\xef\x87\x8a"
#define ICON_FA_VK "\xef\x86\x89"
#define ICON_FA_VNV "\xef\x90\x8b"
#define ICON_FA_VUEJS "\xef\x90\x9f"
#define ICON_FA_WAZE "\xef\xa0\xbf"
#define ICON_FA_WEEBLY "\xef\x97\x8c"
#define ICON_FA_WEIBO "\xef\x86\x8a"
#define ICON_FA_WEIXIN "\xef\x87\x97"
#define ICON_FA_WHATSAPP "\xef\x88\xb2"
#define ICON_FA_WHATSAPP_SQUARE "\xef\x90\x8c"
#define ICON_FA_WHMCS "\xef\x90\x8d"
#define ICON_FA_WIKIPEDIA_W "\xef\x89\xa6"
#define ICON_FA_WINDOWS "\xef\x85\xba"
#define ICON_FA_WIX "\xef\x97\x8f"
#define ICON_FA_WIZARDS_OF_THE_COAST "\xef\x9c\xb0"
#define ICON_FA_WOLF_PACK_BATTALION "\xef\x94\x94"
#define ICON_FA_WORDPRESS "\xef\x86\x9a"
#define ICON_FA_WORDPRESS_SIMPLE "\xef\x90\x91"
#define ICON_FA_WPBEGINNER "\xef\x8a\x97"
#define ICON_FA_WPEXPLORER "\xef\x8b\x9e"
#define ICON_FA_WPFORMS "\xef\x8a\x98"
#define ICON_FA_WPRESSR "\xef\x8f\xa4"
#define ICON_FA_XBOX "\xef\x90\x92"
#define ICON_FA_XING "\xef\x85\xa8"
#define ICON_FA_XING_SQUARE "\xef\x85\xa9"
#define ICON_FA_Y_COMBINATOR "\xef\x88\xbb"
#define ICON_FA_YAHOO "\xef\x86\x9e"
#define ICON_FA_YAMMER "\xef\xa1\x80"
#define ICON_FA_YANDEX "\xef\x90\x93"
#define ICON_FA_YANDEX_INTERNATIONAL "\xef\x90\x94"
#define ICON_FA_YARN "\xef\x9f\xa3"
#define ICON_FA_YELP "\xef\x87\xa9"
#define ICON_FA_YOAST "\xef\x8a\xb1"
#define ICON_FA_YOUTUBE "\xef\x85\xa7"
#define ICON_FA_YOUTUBE_SQUARE "\xef\x90\xb1"
#define ICON_FA_ZHIHU "\xef\x98\xbf"
