
/*
 * Copyright (c) 2021
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef _MACPORTS_SYSRANDOM_H_
#define _MACPORTS_SYSRANDOM_H_

/* MP support header */
#include "MacportsLegacySupport.h"

/* Include the primary system random.h */
#include_next <sys/random.h>

#if __MP_LEGACY_SUPPORT_GETENTROPY__

__MP__BEGIN_DECLS

extern int getentropy(void *buf, size_t buflen);

__MP__END_DECLS

#endif /* __MP_LEGACY_SUPPORT_GETENTROPY__ */

#endif /* _MACPORTS_SYSRANDOM_H_ */
